//-----------------------------------------------------------
//    synth_controller.cpp
//-----------------------------------------------------------
#include "synth_controller.h"
#include "signal_process_base.h"
#include "vco.h"
#include "vcf.h"
#include "vca.h"
#include "to_portaudio.h"
#include "mixer.h"
//-----------------------------------------------------------
//    const
//-----------------------------------------------------------
const int id_topa = 0;
const int id_vco1 = 1;
const int id_vco2 = 2;
const int id_mixer = 3;
const int id_vcf = 4;
const int id_vca = 5;
//-----------------------------------------------------------
//    constructor
//-----------------------------------------------------------
SynthController::SynthController()
{
	SetSynthParts();
}

//-----------------------------------------------------------
//    destructor
//-----------------------------------------------------------
SynthController::~SynthController()
{
	StopSynth();
	for(int i1 = 0, sz = synthparts.size(); i1 < sz; i1++){
		delete synthparts[i1];
	}
	synthparts.clear();
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthController::NoteOn(const float freq)
{
	int i1, sz = synthparts.size(), cnt;
	SignalProcessType sptype;
	ToPortAudio* pa1 = (ToPortAudio*)GetSynthPartById(id_topa);
	if(pa1==0) return;
	cnt = pa1->GetPhase();
	
	for(i1 = 0; i1 < sz; i1++){
		sptype = synthparts[i1]->GetType();
		switch(sptype){
			case VCO:
				{
					SynthVCO* vcotmp = (SynthVCO*)synthparts[i1];
					vcotmp->SetFrequency(freq);
					synthparts[i1]->SetNoteOnCount(cnt);
				}
				break;
			case VCF:
				{
					SynthVCF* vcftmp = (SynthVCF*)synthparts[i1];
					vcftmp->SetCutOffFrequency(freq*2.0);
					synthparts[i1]->SetNoteOnCount(cnt);
				}
				break;
			case VCA:
				synthparts[i1]->SetNoteOnCount(cnt);
				break;
			default:
				break;
		}
	}
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthController::NoteOff()
{
	int i1, sz = synthparts.size();
	SignalProcessType sptype;
	ToPortAudio* pa1 = (ToPortAudio*)GetSynthPartById(id_topa);
	if(pa1==0) return;
	int cnt = pa1->GetPhase();
	for(i1 = 0; i1 < sz; i1++){
		sptype = synthparts[i1]->GetType();
		switch(sptype){
			case VCO:
			case VCF:
			case VCA:
				synthparts[i1]->SetNoteOffCount(cnt);
				break;
			default:
				break;
		}
	}
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthController::StartSynth()
{
	int i1, sz = synthparts.size();
	SignalProcessType sptype;
	for(i1 = 0; i1 < sz; i1++){
		synthparts[i1]->StartRoutine();
	}
	
	for(i1 = 0; i1 < sz; i1++){
		sptype = synthparts[i1]->GetType();
		switch(sptype){
			case VCO:
				synthparts[i1]->SetTask();
				break;
			default:
				break;
		}
	}
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthController::StopSynth()
{
	int i1, sz = synthparts.size();
	ToPortAudio* patmp = (ToPortAudio*)synthparts[0];
	patmp->stop_pa();
	for(i1 = 0; i1 < sz; i1++){
		synthparts[i1]->AccomplishTask();
	}
	for(i1 = 0; i1 < sz; i1++){
		synthparts[i1]->StopRoutine();
	}
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
const SignalProcessBase* SynthController::GetSynthPart(int n1) const
{
	if((n1 < 0) || (n1 >= synthparts.size())) return 0;
	return synthparts[n1];
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
const SignalProcessBase* SynthController::GetSynthPartById(int id1) const
{
	for(int i1 = 0, sz = synthparts.size(); i1 < sz; i1++){
		if(synthparts[i1]->GetId()==id1) return synthparts[i1];
	}
	return 0;
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthController::SetSynthParts()
{
	SynthVCO *vco1, *vco2;
	SynthVCF *vcf1;
	SynthVCA *vca1;
	Mixer *mixer1;
	ToPortAudio *to_pa1;
	std::vector<SignalProcessBase*> vcos;
	
	vco1 = new SynthVCO(id_vco1);
	vco2 = new SynthVCO(id_vco2);
	mixer1 = new Mixer(id_mixer);
	vcf1 = new SynthVCF(id_vcf);
	vca1 = new SynthVCA(id_vca);
	to_pa1 = new ToPortAudio(id_topa);
	
	vcos.push_back(vco1);
	vcos.push_back(vco2);
	vco1->SetUpStream((SignalProcessBase*)to_pa1);
	vco1->SetDownStream((SignalProcessBase*)mixer1);
	vco2->SetUpStream((SignalProcessBase*)to_pa1);
	vco2->SetDownStream((SignalProcessBase*)mixer1);
	mixer1->SetUpStream(vcos);
	mixer1->SetDownStream((SignalProcessBase*)vcf1);
	vcf1->SetUpStream((SignalProcessBase*)mixer1);
	vcf1->SetDownStream((SignalProcessBase*)vca1);
	vca1->SetUpStream((SignalProcessBase*)vcf1);
	vca1->SetDownStream((SignalProcessBase*)to_pa1);
	to_pa1->SetUpStream((SignalProcessBase*)vca1);
	to_pa1->SetDownStream(vcos);
	
	
	vco1->SetWaveType(SAW);
	
	synthparts.push_back(to_pa1);
	synthparts.push_back(vco1);
	synthparts.push_back(vco2);
	synthparts.push_back(mixer1);
	synthparts.push_back(vcf1);
	synthparts.push_back(vca1);
}

