//-----------------------------------------------------------
//    vcf.cpp
//-----------------------------------------------------------
#include "vcf.h"
#include "eg.h"
#include <math.h>
//-----------------------------------------------------------
//    
//-----------------------------------------------------------
const float default_frequency = 2000.0;
const float SynthVCF::min_freq = 100.0;
const float SynthVCF::max_freq = 10000.0;
const unsigned int SynthVCF::bw_table_size = 20000;
const float SynthVCF::delta_freq
	= (10000.0 - 100.0) / 20000.0;
float** SynthVCF::butterworth4_ai = 0;
float** SynthVCF::butterworth4_bi = 0;
//-----------------------------------------------------------
//    
//-----------------------------------------------------------
SynthVCF::SynthVCF(int id1)
	: SignalProcessBase(id1)
{
	SetSignalProcessType(VCF);
	eg1 = new SynthEG();
	SetTable();
	SetCutOffFrequency(default_frequency);
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
SynthVCF::~SynthVCF()
{
	delete eg1;
	if(!ExistOtherVCF()){
		for(int i1 = 0; i1 < bw_table_size; i1++){
			delete butterworth4_ai[i1];
			delete butterworth4_bi[i1];
		}
		delete [] butterworth4_ai;
		delete [] butterworth4_bi;
		butterworth4_ai = 0;
		butterworth4_bi = 0;
	}
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthVCF::Work()
{
	static float cbuf_inp[4] = {0.0, 0.0, 0.0, 0.0};
	static float cbuf_out[4] = {0.0, 0.0, 0.0, 0.0};
	unsigned int bsz = GetSizeBuffer();
	float itmp[bsz+4], itmp2[bsz];
	float otmp[bsz+4], otmp2[bsz];
	float f1;
	GetUpStream(0)->CopyBuffer(itmp2);
	int oncnt = GetNoteOnCount(), offcnt = GetNoteOffCount();
	
	for(int i1 = 0 ; i1 < 4; i1++){
		itmp[i1] = cbuf_inp[i1];
		otmp[i1] = cbuf_out[i1];
	}
	for(int i1 = 0; i1 < bsz; i1++){
		itmp[i1+4] = itmp2[i1];
	}
	if(oncnt >= 0)
	{
		for(int i1 = 0; i1 < bsz; i1++){
			if(i1 < oncnt){
				otmp[i1+4] = itmp[i1+4];
			}else if(i1 == oncnt){
				eg1->SetEGState(ATTACK);
				SetNoteOnCount(-1);
				f1 = cutoff * 0.5 + (cutoff * 0.5) * eg1->GetValue();
				SetButterworthConstants(f1);
				otmp[i1+4] = ai[0] * itmp[i1] + ai[1] * itmp[i1+1]
					+ ai[2] * itmp[i1+2] + ai[3] * itmp[i1+3] + ai[4] * itmp[i1+4];
				otmp[i1+4] -= bi[0] * otmp[i1] + bi[1] * otmp[i1+1]
					+ bi[2] * otmp[i1+2] + bi[3] * otmp[i1+3];
			}else if(i1 > oncnt){
				if(i1==offcnt){
					eg1->SetEGState(RELEASE);
					SetNoteOffCount(-1);
				}
				f1 = cutoff * 0.5 + (cutoff * 0.5) * eg1->GetValue();
				SetButterworthConstants(f1);
				otmp[i1+4] = ai[0] * itmp[i1] + ai[1] * itmp[i1+1]
					+ ai[2] * itmp[i1+2] + ai[3] * itmp[i1+3] + ai[4] * itmp[i1+4];
				otmp[i1+4] -= bi[0] * otmp[i1] + bi[1] * otmp[i1+1]
					+ bi[2] * otmp[i1+2] + bi[3] * otmp[i1+3];
			}
		}
		for(int i1 = 0; i1 < 4; i1++){
			cbuf_inp[i1] = itmp[bsz + i1];
			cbuf_out[i1] = otmp[bsz + i1];
		}
		for(int i1 = 0; i1 < bsz; i1++){
			otmp2[i1] = otmp[i1+4];
		}
		WriteBuffer(otmp2);
		
	}
	else if((eg1->GetEGState() != NON))
	{
		for(int i1 = 0; i1 < bsz; i1++){
			if(i1==offcnt){
				eg1->SetEGState(RELEASE);
				SetNoteOffCount(-1);
			}
			f1 = cutoff * 0.5 + (cutoff * 0.5) * eg1->GetValue();
			SetButterworthConstants(f1);
			otmp[i1+4] = ai[0] * itmp[i1] + ai[1] * itmp[i1+1]
				+ ai[2] * itmp[i1+2] + ai[3] * itmp[i1+3] + ai[4] * itmp[i1+4];
			otmp[i1+4] -= bi[0] * otmp[i1] + bi[1] * otmp[i1+1]
				+ bi[2] * otmp[i1+2] + bi[3] * otmp[i1+3];
		}
		for(int i1 = 0; i1 < 4; i1++){
			cbuf_inp[i1] = itmp[bsz + i1];
			cbuf_out[i1] = otmp[bsz + i1];
		}
		for(int i1 = 0; i1 < bsz; i1++){
			otmp2[i1] = otmp[i1+4];
		}
		WriteBuffer(otmp2);
		
	}
	else
	{
		for(int i1 = 0; i1 < 4; i1++){
			cbuf_inp[i1] = 0.0;
			cbuf_out[i1] = 0.0;
		}
		InitializeBuffer();
	}
	
	GiveTask();
	AccomplishTask();
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthVCF::SetCutOffFrequency(const float frequency)
{
	if(frequency < 0.0) return;
	cutoff = frequency;
}


//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthVCF::SetButterworthConstants(const float frequency)
{
	int n1;
	float c1;
	if(frequency < min_freq){
		n1 = 0;
	}else if(frequency > max_freq){
		n1 = bw_table_size - 1;
	}else{
		c1 = (frequency - min_freq) / delta_freq;
		n1 = (int)c1;
	}
	
	ai[0] = butterworth4_ai[n1][0];
	ai[1] = butterworth4_ai[n1][1];
	ai[2] = butterworth4_ai[n1][2];
	ai[3] = butterworth4_ai[n1][3];
	ai[4] = butterworth4_ai[n1][4];
	bi[0] = butterworth4_bi[n1][0];
	bi[1] = butterworth4_bi[n1][1];
	bi[2] = butterworth4_bi[n1][2];
	bi[3] = butterworth4_bi[n1][3];
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthVCF::SetTable()
{
	if((butterworth4_ai == 0)||(butterworth4_bi == 0)){
		delete [] butterworth4_ai;
		delete [] butterworth4_bi;
		butterworth4_ai = new float*[bw_table_size];
		butterworth4_bi = new float*[bw_table_size];
		for(int i1 = 0; i1 < bw_table_size; i1++){
			butterworth4_ai[i1] = new float[5];
			butterworth4_bi[i1] = new float[4];
			butterworth4_ai[i1][0] = 0.0;
			butterworth4_ai[i1][1] = 0.0;
			butterworth4_ai[i1][2] = 0.0;
			butterworth4_ai[i1][3] = 0.0;
			butterworth4_ai[i1][4] = 0.0;
			butterworth4_bi[i1][0] = 0.0;
			butterworth4_bi[i1][1] = 0.0;
			butterworth4_bi[i1][2] = 0.0;
			butterworth4_bi[i1][3] = 0.0;
		}
		
		float freq = min_freq;
		for(int i1 = 0; i1 < bw_table_size; i1++){
			CalcButterWorthConstants(freq, butterworth4_ai[i1], butterworth4_bi[i1]);
			freq += delta_freq;
		}
	}
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
void SynthVCF::CalcButterWorthConstants(float frequency, float a[5], float b[4])
{
	float sample_rate = (float)GetSampleRate();
	float wd;
	wd = 2.0 * (float)sample_rate * tan(frequency * M_PI / (float)sample_rate);
	float dt = wd / (float)sample_rate;
	float dt2 = dt * dt;
	float dt3 = dt2 * dt;
	float dt4 = dt2 * dt2;
	float ci[3] = {2.6132, 3.41430612, 2.6132};
	float b0 = 1.0 / (dt4 + 2.0 * dt3 * ci[2]
		+ 4.0 * dt2 * ci[1] + 8.0 * dt * ci[0] + 16.0);
	b[0] = (dt4 - 2.0 * dt3 * ci[2]
		+ 4.0 * dt2 * ci[1] - 8.0 * dt * ci[0] + 16.0) * b0;
	b[1] = (4.0 * dt4 - 4.0 * dt3 * ci[2]
		+ 16.0 * dt * ci[0] - 64.0) * b0;
	b[2] = (6.0 * dt4 - 8.0 * dt2 * ci[1] + 96.0) * b0;
	b[3] = (4.0 * dt4 + 4.0 * dt3 * ci[2]
		- 16.0 * dt * ci[0] - 64.0) * b0;
	float a0 = dt4 * b0;
	a[0] = a0;
	a[1] = 4.0 * a0;
	a[2] = 6.0 * a0;
	a[3] = 4.0 * a0;
	a[4] = a0;
}

//-----------------------------------------------------------
//    
//-----------------------------------------------------------
bool SynthVCF::ExistOtherVCF() const
{
	int spb_sz = GetSignalProcesses().size();
	int idtmp, thisid = this->GetId();
	for(int i1 = 0; i1 < spb_sz; i1++){
		idtmp = GetSignalProcesses()[i1]->GetId();
		if(idtmp!=thisid){
			if(GetSignalProcesses()[i1]->GetType()==VCF) return true;
		}
	}
	return false;
}

